///////////// Copyright © 2016 Fabian Oboril. All rights reserved. /////////
//
//   Project     : Altera OpenCL Kernels
//   File        : laplace.cl
//   Description :
//      out = A*x, where A is the laplace matrix
//
//   Created On: 09.09.2016
//   Created By: Fabian Oboril
////////////////////////////////////////////////////////////////////////////

/**
 * Basic kernel computes all entries using float16
 */
__kernel
void kernel_laplace_basic( __global const float16 *x,  __global float16 *out , const int size, const int nx) {

    int gid = get_global_id(0);

    float16 n = (gid < size-nx)                       ? x[gid+nx] : (float16)(0);
    float16 s = (gid > nx-1)                          ? x[gid-nx] : (float16)(0);
    float16 w = (((gid % nx) && (gid > 0)) == 0)      ? (float16)(        0,x[gid][0],x[gid][1],x[gid][2],
                x[gid][3],x[gid][4],x[gid][5],x[gid][6],
                x[gid][7],x[gid][8],x[gid][9],x[gid][10],
                x[gid][11],x[gid][12],x[gid][13],x[gid][14]) :
                (float16)(x[gid-1][15],x[gid][0],x[gid][1],x[gid][2],
                          x[gid][3],x[gid][4],x[gid][5],x[gid][6],
                          x[gid][7],x[gid][8],x[gid][9],x[gid][10],
                          x[gid][11],x[gid][12],x[gid][13],x[gid][14]);
    float16 e = (((gid+1) % nx) == 0)                 ? (float16)(x[gid][1],x[gid][2],x[gid][3],x[gid][4],
                x[gid][5],x[gid][6],x[gid][7],x[gid][8],
                x[gid][9],x[gid][10],x[gid][11],x[gid][12],
                x[gid][13],x[gid][14],x[gid][15],0) :
                (float16)(x[gid][1],x[gid][2],x[gid][3],x[gid][4],
                          x[gid][5],x[gid][6],x[gid][7],x[gid][8],
                          x[gid][9],x[gid][10],x[gid][11],x[gid][12],
                          x[gid][13],x[gid][14],x[gid][15],x[gid+1][0]);

    out[gid] = 4*x[gid] - n - s - w - e;
}



/**
 * In the advanced kernels the if-else conditions for boundary elements are avoided by splitting the kernels
 */
__kernel
void kernel_laplace_advanced_center( __global float16 * restrict x,  __global float16 * restrict out , const int size, const int nx) {

    int gid = (((get_global_id(0)*4)/nx*2)*nx + (get_global_id(0)*4)%nx);

    float16 out_local[4];
    float16 m[4] = {x[gid],x[gid+1],x[gid+2],x[gid+3]};
    float16 n[4] = {x[gid+nx],x[gid+nx+1],x[gid+nx+2],x[gid+nx+3]};
    float16 s[4] = {x[gid-nx],x[gid+nx+1],x[gid+nx+2],x[gid+nx+3]};

    float16 w[4];
    w[0] = (float16)(x[gid-1][15], m[0][0], m[0][1], m[0][2], m[0][3], m[0][4], m[0][5], m[0][6], m[0][7], m[0][8], m[0][9], m[0][10],m[0][11], m[0][12], m[0][13], m[0][14]);
    w[1] = (float16)(m[0][15], m[1][0], m[1][1], m[1][2], m[1][3], m[1][4], m[1][5], m[1][6], m[1][7], m[1][8], m[1][9], m[1][10],m[1][11], m[1][12], m[1][13], m[1][14]);
    w[2] = (float16)(m[1][15], m[2][0], m[2][1], m[2][2], m[2][3], m[2][4], m[2][5], m[2][6], m[2][7], m[2][8], m[2][9], m[2][10],m[2][11], m[2][12], m[2][13], m[2][14]);
    w[3] = (float16)(m[2][15], m[3][0], m[3][1], m[3][2], m[3][3], m[3][4], m[3][5], m[3][6], m[3][7], m[3][8], m[3][9], m[3][10],m[3][11], m[3][12], m[3][13], m[3][14]);

    float16 e[4];
    e[0] = (float16)(m[0][1], m[0][2], m[0][3], m[0][4], m[0][5], m[0][6], m[0][7], m[0][8], m[0][9], m[0][10],m[0][11], m[0][12], m[0][13], m[0][14], m[0][15], m[1][0]);
    e[1] = (float16)(m[1][1], m[1][2], m[1][3], m[1][4], m[1][5], m[1][6], m[1][7], m[1][8], m[1][9], m[1][10],m[1][11], m[1][12], m[1][13], m[1][14], m[1][15], m[2][0]);
    e[2] = (float16)(m[2][1], m[2][2], m[2][3], m[2][4], m[2][5], m[2][6], m[2][7], m[2][8], m[2][9], m[2][10],m[2][11], m[2][12], m[2][13], m[2][14], m[2][15], m[3][0]);
    e[3] = (float16)(m[3][1], m[3][2], m[3][3], m[3][4], m[3][5], m[3][6], m[3][7], m[3][8], m[3][9], m[3][10],m[3][11], m[3][12], m[3][13], m[3][14], m[3][15], x[gid+4][0]);

#pragma unroll
    for (int i=0; i<4; i++) {
        out_local[i] = 4*m[i] - n[i] - s[i] - e[i] - w[i];
        out[gid+i] = out_local[i];
    }


#pragma unroll
    for (int i=0; i<4; i++) {
        s[i] = m[i];
        m[i] = n[i];
        n[i] = x[gid+nx+nx+i];
    }

    w[0] = (float16)(x[gid-1+nx][15], m[0][0], m[0][1], m[0][2], m[0][3], m[0][4], m[0][5], m[0][6], m[0][7], m[0][8], m[0][9], m[0][10],m[0][11], m[0][12], m[0][13], m[0][14]);
    w[1] = (float16)(m[0][15], m[1][0], m[1][1], m[1][2], m[1][3], m[1][4], m[1][5], m[1][6], m[1][7], m[1][8], m[1][9], m[1][10],m[1][11], m[1][12], m[1][13], m[1][14]);
    w[2] = (float16)(m[1][15], m[2][0], m[2][1], m[2][2], m[2][3], m[2][4], m[2][5], m[2][6], m[2][7], m[2][8], m[2][9], m[2][10],m[2][11], m[2][12], m[2][13], m[2][14]);
    w[3] = (float16)(m[2][15], m[3][0], m[3][1], m[3][2], m[3][3], m[3][4], m[3][5], m[3][6], m[3][7], m[3][8], m[3][9], m[3][10],m[3][11], m[3][12], m[3][13], m[3][14]);

    e[0] = (float16)(m[0][1], m[0][2], m[0][3], m[0][4], m[0][5], m[0][6], m[0][7], m[0][8], m[0][9], m[0][10],m[0][11], m[0][12], m[0][13], m[0][14], m[0][15], m[1][0]);
    e[1] = (float16)(m[1][1], m[1][2], m[1][3], m[1][4], m[1][5], m[1][6], m[1][7], m[1][8], m[1][9], m[1][10],m[1][11], m[1][12], m[1][13], m[1][14], m[1][15], m[2][0]);
    e[2] = (float16)(m[2][1], m[2][2], m[2][3], m[2][4], m[2][5], m[2][6], m[2][7], m[2][8], m[2][9], m[2][10],m[2][11], m[2][12], m[2][13], m[2][14], m[2][15], m[3][0]);
    e[3] = (float16)(m[3][1], m[3][2], m[3][3], m[3][4], m[3][5], m[3][6], m[3][7], m[3][8], m[3][9], m[3][10],m[3][11], m[3][12], m[3][13], m[3][14], m[3][15], x[gid+4+nx][0]);

#pragma unroll
    for (int i=0; i<4; i++) {
        out_local[i] = 4*m[i] - n[i] - s[i] - e[i] - w[i];
        out[gid+i+nx] = out_local[i];
    }
}


__kernel
void kernel_laplace_advanced_top_bottom( __global float8 * restrict x,  __global float8 * restrict out , const int size, const int nx) {

    int gid = get_global_id(0);
    int offset = 0;
    float8 out_local, m, n, s, w, e;

    for (int i=0; i<=1; i++) {
        m = x[gid + offset];
        n = x[gid + nx + offset];
        s = x[gid - nx + offset];
        w = (float8)(x[gid-1+ offset][7],m[0],m[1],m[2],
                     m[3],m[4],m[5],m[6]);
        e = (float8)(m[1],m[2],m[3],m[4],
                     m[5],m[6],m[7],x[gid+1+ offset][0]);
        out_local = 4*m - w - e - (1-i)*n - i*s;
        out[gid + offset] = out_local;
        offset = size - nx;
    }
}

__kernel
void kernel_laplace_advanced_left_right( __global float * restrict x,  __global float * restrict out , const int size, const int nx) {

    int gid = get_global_id(0)*nx;

    float out_local, m, n, s, e, w;
    int offset = 0;

    for (int i=0; i<=1; i++) {
        m = x[gid + offset];
        n = (size - nx - gid) ? x[gid+nx+offset] : 0;
        s = (gid) ? x[gid-nx+offset] : 0;
        e = x[gid + 1];
        w = x[gid + offset - 1];
        out_local = 4*m - s - n - (1-i)*e - i*w;
        out[gid + offset] = out_local;
        offset = nx - 1;
    }
}